#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include "fileserv.h"
#include "wordsplit.h"

char *dotfile = ".fileserv";

DIRCONFIG *
dirconfig(char const *path, size_t prefix_len)
{
	DIRCONFIG *conf = dirconfig_init();
	char *tmp = NULL;
	struct stat st;

	if (stat(path, &st) == 0 && S_ISDIR(st.st_mode)) {
		tmp = catfile(path, ".");
		path = tmp;
	}
	while (path[prefix_len]) {
		char *name = catfile_n(path, prefix_len, dotfile);
		dirconfig_parse(name, conf);
		free(name);
		if (path[prefix_len] == '/')
			prefix_len++;
		while (path[prefix_len] && path[prefix_len] != '/')
			prefix_len++;
	}
	free(tmp);
	return conf;
}

static int
bool_decode(char const *s)
{
	static char *yes[] = {
		"1", "yes", "true", "on", NULL
	};
	static char *no[] = {
		"0", "no", "false", "off", NULL
	};
	int i;
	for (i = 0; yes[i]; i++) {
		if (strcasecmp(s, yes[i]) == 0)
			return 1;
	}
	for (i = 0; no[i]; i++) {
		if (strcasecmp(s, no[i]) == 0)
			return 0;
	}
	return -1;
}

static int
bad_argc(const char *file, int line, char const *id)
{
	error("%s:%d: bad number of arguments to '%s'", file, line, id);
	return 1;
}

static int
not_a_bool(const char *file, int line)
{
	error("%s:%d: invalid boolean value", file, line);
	return 1;
}

static int
set_follow(size_t argc, char **argv, DIRCONFIG *conf,
	   char const *file, int line)
{
	int n;
	if (argc != 2)
		return bad_argc(file, line, argv[0]);
	n = bool_decode(argv[1]);
	if (n == -1)
		return not_a_bool(file, line);
	conf->follow = n;
	return 0;
}

static int
set_listing(size_t argc, char **argv, DIRCONFIG *conf,
	    char const *file, int line)
{
	int n;
	if (argc != 2)
		return bad_argc(file, line, argv[0]);
	n = bool_decode(argv[1]);
	if (n == -1)
		return not_a_bool(file, line);
	conf->listing = n;
	return 0;
}

static int
set_list_unreadable(size_t argc, char **argv, DIRCONFIG *conf,
		    char const *file, int line)
{
	int n;
	if (argc != 2)
		return bad_argc(file, line, argv[0]);
	n = bool_decode(argv[1]);
	if (n == -1)
		return not_a_bool(file, line);
	conf->list_unreadable = n;
	return 0;
}

static int
set_index_files(size_t argc, char **argv, DIRCONFIG *conf,
		char const *file, int line)
{
	size_t i;
	
	if (argc < 2)
		return bad_argc(file, line, argv[0]);
	conf->index_files = xcalloc(argc, sizeof(conf->index_files[0]));
	for (i = 1; i < argc; i++)
		conf->index_files[i-1] = xstrdup(argv[i]);
	conf->index_files[i-1] = NULL;
	return 0;
}

static void
hidden_files_free(DIRCONFIG *conf)
{
	size_t i;

	for (i = 0; i < conf->hidden_files_rxc; i++)
		regfree(&conf->hidden_files_rxv[i]);
	conf->hidden_files_rxc = 0;
	free(conf->hidden_files_rxv);
	conf->hidden_files_rxv = NULL;
}
		
static int
set_hidden_files(size_t argc, char **argv, DIRCONFIG *conf,
		char const *file, int line)
{
	size_t i;
	
	if (argc < 2)
		return bad_argc(file, line, argv[0]);
	
	i = 1;
	if (strcmp(argv[i], "+") == 0)
		i++;
	else
		hidden_files_free(conf);

	conf->hidden_files_rxv = xnrealloc(conf->hidden_files_rxv,
					   argc + conf->hidden_files_rxc,
					   sizeof(conf->hidden_files_rxv[0]));
	for (; argv[i]; i++) {
		int ec;
		ec = regcomp(&conf->hidden_files_rxv[conf->hidden_files_rxc],
			     argv[i], REG_EXTENDED|REG_NOSUB);
		if (ec) {
			char buf[512];
			regerror(ec,
				 &conf->hidden_files_rxv[conf->hidden_files_rxc],
				 buf, sizeof buf);
			error("%s: %s", argv[i], buf);
		} else
			conf->hidden_files_rxc++;
	}
	return 0;
}

struct dirconfig_keyword {
	char const *ident;
	int (*setter)(size_t argc, char **argv, DIRCONFIG *conf,
		      char const *file, int line);
};
	
static struct dirconfig_keyword keywords[] = {
	{ "directory-index", set_index_files },
	{ "follow", set_follow },
	{ "listing", set_listing },
	{ "list-unreadable", set_list_unreadable },
	{ "hidden-files", set_hidden_files },
	{ NULL }
};

static int
line_interpret(struct wordsplit const *ws,
	       DIRCONFIG *conf, char const *file, int line)
{
	struct dirconfig_keyword *kw;
	for (kw = keywords; kw->ident; kw++) {
		if (strcmp(kw->ident, ws->ws_wordv[0]) == 0)
			return kw->setter(ws->ws_wordc, ws->ws_wordv,
					  conf, file, line);
	}
	error("%s:%d: unrecognized keyword", file, line);
	return 1;
}

void
dirconfig_parse(const char *file, DIRCONFIG *conf)
{
	FILE *fp;
	char buf[1024];
	unsigned ln = 0;
	int wsflags = WRDSF_DEFFLAGS;
	struct wordsplit ws;

	fp = fopen(file, "r");
	if (!fp) {
		if (errno != ENOENT)
			error("can't open file %s: %s", file,
			      strerror(errno));
		return;
	}

	ws.ws_comment = "#";
	wsflags |= WRDSF_COMMENT;
	
	while (fgets(buf, sizeof(buf), fp)) {
		int len;

		ln++;
		len = strlen(buf);
		if (len == 0)
			continue;
		if (buf[len-1] == '\n')
			buf[--len] = 0;
		else if (!feof(fp)) {
			error("%s:%d: line too long", file, ln);
			break;
		}

		if (wordsplit(buf, &ws, wsflags)) {
			error("%s:%d: %s", file, ln, wordsplit_strerror(&ws));
			break;
		}

		wsflags |= WRDSF_REUSE;
		if (ws.ws_wordc == 0)
			continue;
		if (line_interpret(&ws, conf, file, ln)) {
			break;
		}
	}

	if (wsflags & WRDSF_REUSE)
		wordsplit_free(&ws);

	if (ferror(fp)) {
		error("%s: %s", file, strerror(errno));
	}
	fclose(fp);
}

DIRCONFIG *
dirconfig_init(void)
{
	DIRCONFIG *p = xcalloc(1, sizeof(*p));
	p->list_unreadable = 1;
	return p;
}

void
dirconfig_free(DIRCONFIG *conf)
{
	if (conf) {
		if (conf->index_files) {
			size_t i;
			for (i = 0; conf->index_files[i]; i++)
				free(conf->index_files[i]);
			free(conf->index_files);
		}
		hidden_files_free(conf);
		free(conf);
	}
}

int
filename_is_valid(char const *name)
{
	enum { a_chr, a_sla, a_dot, a_end, A_MAX };
	enum { s_ini, s_cur, s_sla, s_dot, s_dt2, s_err, s_end, S_MAX = s_end }
	     state = s_ini;
	static char trans[A_MAX][S_MAX] = {
		/*           ini     cur    sla    dot    dt2    err */
		/* chr */ { s_cur,  s_cur, s_cur, s_cur, s_cur, s_err },  
		/* sla */ { s_sla,  s_sla, s_err, s_err, s_err, s_err },
		/* dot */ { s_dot,  s_cur, s_dot, s_dt2, s_cur, s_err },
		/* end */ { s_end,  s_end, s_end, s_err, s_err, s_err },
	};
	
	int c;
	while (state != s_err && state != s_end) {
		switch (*name++) {
		case '/':
			c = a_sla;
			break;
		case '.':
			c = a_dot;
			break;
		case 0:
			c = a_end;
			break;
		default:
			c = a_chr;
		}
		state = trans[c][state];
	}
	return state == s_end;
}
			
int
filename_is_special(char const *name)
{
	return (name[0] == '.'
		&& (name[1] == 0
		    || (name[1] == '.' && name[2] == 0)));
}

int
filename_is_hidden(char const *name, DIRCONFIG const *conf)
{
	if (filename_is_special(name) || strcmp(name, dotfile) == 0)
		return 1;

	if (conf && conf->hidden_files_rxv) {
		size_t i;
		for (i = 0; i < conf->hidden_files_rxc; i++)
			if (regexec(&conf->hidden_files_rxv[i], name,
				    0, NULL, 0) == 0)
				return 1;
	}
	return 0;
}

		
			
